/* par_act_net.c   */
/*
   Copyright (C) 2006 John W. Jordan
   All Rights Reserved
*/
/* 
  This file is part of Dante.
  Dante is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the license, or
  ( at your option ) any later version.
  Dante is distributed in the hope that it will be usefull,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A OARTICULAR PURPOSE. See the
  GNU General Public License for more detaile.
  You should have received a copy of the GNU General Public License
  along with Dante; if not write to the Free Software
  Foundation. Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
USA
*/

/* Generates a syntax net as operators are encountered.
   It mimics processing a postfix string on a stack.
   Operators are processed immediately instead of actually 
   constructing a postscript string.
*/
#include <stdlib.h>
#include "par.h"
#include <ctype.h>
/* --------------------------------------------------------- */
void pr_stacks (PAR * par, char *label);
static void join (NODE * left, NODE * right);
static void set_bt (NODE * node, NODE * right);
static void set_bf (NODE * node, NODE * right);
static void set_error (NODE * node);
static void set_repeat (NODE * node, NODE * loop);
/* --------------------------------------------------------- */
static const int debug = 0;
/* ----------------- AND ----------------- */
int
par_a_and (PAR * par)
{
  NODE *left;
  NODE *right;
  NODE *x;

  right = (NODE *) Pop (par->stk_opnd);
  left = (NODE *) Pop (par->stk_opnd);

  if ((left == 0) || (right == 0))
    {
      fprintf (stderr, "NET: AND op has null opnd.\n");
      exit (1);
    }

  /* Local code is detected here since it is read by the factor rule. */
  if (left->type == 'D')
    {
      Push (par->stk_opnd, left);
      Push (par->stk_opnd, right);
      if (debug)
	printf ("LOCAL CODE: left [%s] right [%s]\n",
		left->name, right->name);
      return;
    }

  /* Branch if true. */
  for (x = left; x; x = x->nlink)
    {
      /* The usual successor case. */
      if (x->suc == (NODE *) 1)
	x->suc = right;
      /* For a node inside {}. */
      if (x->alt == (NODE *) 1)
	x->alt = right;
    }
  /* Set error. */
  for (x = right; x; x = x->nlink)
    if (x->alt == (NODE *) 0)
      x->alt = (NODE *) 2;

  /* Join left subtree to right subtree. */
  for (x = left; x->nlink; x = x->nlink)
    NULL;
  x->nlink = right;

  /* New operand for stack. */
  Push (par->stk_opnd, left);
}

/* ----------------- OR ----------------- */
int
par_a_or (PAR * par)
{
  NODE *left;
  NODE *right;
  NODE *x;

  /* Binary operation. */
  right = (NODE *) Pop (par->stk_opnd);
  left = (NODE *) Pop (par->stk_opnd);

  /* Branch if false. */
  for (x = left; x; x = x->nlink)
    if (x->alt == (NODE *) 0)
      {
	x->alt = right;
	break;
      }

  /* Join left subtree to right subtree. */
  for (x = left; x->nlink; x = x->nlink)
    NULL;
  x->nlink = right;

  /* New operand for stack. */
  Push (par->stk_opnd, left);
}

/* --------------------------------------------------------- */
void
par_a_rpt (PAR * par)
{
  NODE *node;
  NODE *right;
  NODE *x;

  /* Unitary operator. */
  right = (NODE *) Pop (par->stk_opnd);
  if (right == 0)
    {
      fprintf (stderr, "NET: repeat [%s] has null opnd  .\n", right->name);
      exit (1);
    }

  for (x = right; x; x = x->nlink)
    {
      if (x->suc == (NODE *) 1)
	x->suc = right;
      if (x->alt == (NODE *) 0)
	x->alt = (NODE *) 1;

      /* Added for not operator. */
      if (x->suc == (NODE *) 0)
	x->suc = (NODE *) 1;
    }
  Push (par->stk_opnd, right);
}

/* --------------------------------------------------------- */
void
par_a_opt (PAR * par)
{
  NODE *node;
  NODE *right;
  NODE *x;

  right = (NODE *) Pop (par->stk_opnd);
  if (right == 0)
    {
      printf ("NET:opt[%s] has null opnd  .\n", right->name);
      exit (1);
    }
  for (x = right; x; x = x->nlink)
    {
      /* suc can be 0 when not is included in the language. */
      if (x->suc == (NODE *) 0)
	x->suc = (NODE *) 1;
      if (x->alt == (NODE *) 0)
	x->alt = (NODE *) 1;
    }
  Push (par->stk_opnd, right);
}

/* --------------------------------------------------------- */
void
par_a_not (PAR * par)
{
  NODE *right;

  par_a_iden (par);

  right = (NODE *) Pop (par->stk_opnd);
  right->alt = (NODE *)1;
  right->suc = (NODE *)0;
  Push (par->stk_opnd, right);
}
